/*
   Bezier Curve Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to create a Beizer curve in OpenGL.
   Creating a Bezier curve in OpenGL takes a few steps.  These steps
   are in the RenderScene function.  In fact there are only 3 functions you
   need which are in RenderScene().
*/


// Demo: Bezier Curve.


#define WIN32_LEAN_AND_MEAN      // This trims down the libraries.
#define VC_LEANMEAN              // Trims even farther.


// All the necessary includes...
#include<windows.h>              // The windows header file.
#include<gl/glu.h>               // Opengl utilities.
#include<gl/glaux.h>             // Glaux header.


// Function Prototypes...
void RenderScene();              // Renders the scene.
bool InitializeGL();             // Initialize the window.


// This will be used to swap buffers.
HDC g_HDC;

// Here I will define the points.  There are two points for the line
// and two points (control points) to bend the line to make a curve.
// The order is point, control point 1, control point 2, point 2.
float points[4][3] = {{-3.0f, -3.0f, 0.0f}, {-3.0f, 1.5f, 0.0f},
                      {5.0f, -2.0f, 0.0f}, {2.0f, 2.0f, 0.0f}};


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // this chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                      // Rendering context.
   static HDC hDC;                        // Device context.
   int width, height;                     // The window width and height.

   switch(message)
      {
         case WM_CREATE:                  // Windows creation.
            hDC = GetDC(hwnd);            // This gets the device context for our window.
            g_HDC = hDC;                  // Assigns the global device context to this one.
            SetupPixelFormat(hDC);        // Call the pixel format function.

            hRC = wglCreateContext(hDC);  // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);     // Makes the rendering context.
            return 0;
            break;

         case WM_CLOSE:                   // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);        // Deletes the rendering context.

            PostQuitMessage(0);           // Says close the program.
            return 0;
            break;

         case WM_SIZE:                    // re-size message.
            height = HIWORD(lParam);      // This gets the height of the window.
            width = LOWORD(lParam);       // This gets the width of the window.

            if(height==0)                 // we don't want it to be possible for a
               {                          // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height);// resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);    // Sets the projection matrix.
            glLoadIdentity();               // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);     // Sets the projection matrix.
            glLoadIdentity();               // Reset the modelview matrix.

            return 0;
            break;

         case WM_KEYDOWN:
		      switch(wParam)
               { 
		            case VK_ESCAPE: // Close the app if the user presses escape.
				         PostQuitMessage(0);
				         break;
               }
            break;

         default:
            break;
      }
   
   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create the window.
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Bezier Curve - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640, 480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;


   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);  // Translate any messages.
               DispatchMessage(&msg);   // Dispatch any messages.
            }
         else
            {
               RenderScene();          // Render a frame.
            }
      }

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.0, 0.0, 0.0, 1.0);

   glShadeModel(GL_SMOOTH);
   glEnable(GL_DEPTH_TEST);

   return true;
}


void RenderScene()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Clears the screen.
   glLoadIdentity();                   //Reset modelview matrix for new frame.

   // white color.
   glColor3f(1.0, 1.0, 1.0);
 
   // Move the scene before drawing.
   glTranslatef(0.0f, 0.0f, -13.0f);

   // This will set up the Bezier curve.  glMap1f defines a 1D evaluator.
   // The prototype is...
   // void glMap1f(GLenum target, float u1, float u2, int stride,
   //              int order, const float *points).
   
   // target defines the vertex coordinates (ex. x, y, z), the u1 and u2
   // define the range for the u parameter.  Stride defines the distance
   // between each point in the curve. and points is the array of points.
   // Check the MSDN for a better explanation if you seek one.
   glMap1f(GL_MAP1_VERTEX_3, 0.0, 1.0, 3, 4, &points[0][0]);

   // Here we enable the GL_MAP1_VERTEX_3.
   glEnable(GL_MAP1_VERTEX_3);

   // Now lets draw the curve using line strips.  All you need is to
   // call glEvalCoord1f() to draw the curve and thats it.
   glBegin(GL_LINE_STRIP);

      for(int i = 0; i < 100; i++)
         {
            // Here you we are drawing 100 points along the curve to
            // make the line.  To do so you send in glEvalCoord1f
            // and send to it i divided by 100.0f (don't forget to cast).
            glEvalCoord1f((float)i / 100.0f);
         }

   glEnd();

   // And that is it.  Here I will draw the points big so the user (you) can
   // see them and where they are.

   // glPointSize() will increase the size of a point (which is the size of a pixel)
   // to whether size you would like.  This is done here so you can see the points.
   glPointSize(4.0);
   
   glBegin(GL_POINTS);

   for(int j = 0; j < 4; j++)
      {
         // Instead of using glVertex3f (which takes in three values) we use
         // glVertex3fv which takes three dimensional arrays.
         glVertex3fv(&points[j][0]);
      }

   glEnd();

   SwapBuffers(g_HDC);
}


// Recap:

// Now that you know how to create a curve (or at least have seen it done)
// you can now move on to curve surfaces then from there the nurbs.
// If you don't understand the 3 functions used to create the
// curve then play with the values to see the different results.  That
// should help clear things up for you.


// Copyright June 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com